"""
spotilyzer init subcommand
"""

# system imports
import os
import shutil

# project imports
from .base import SubCommand
from .utils.paths import get_package_datadir, get_user_datadir, SEEDS_FILE

# constants
_DESCRIPTION = "initialize spotilyzer"
_SEEDS_FILE = 'seeds.json'


class Init(SubCommand):
    """
    init subcommand
    """

    name = 'init'

    @classmethod
    def add_parser(cls, subparsers):
        """
        Add init subcommand parser.
        :param subparsers: object to attach parser
        :return: None
        """
        subparsers.add_parser(cls.name, description=_DESCRIPTION,
                              help=_DESCRIPTION)

    @staticmethod
    def run():
        """
        Copy package seeds file so user can edit it.
        :return: None
        """
        spotilyzer_dir = get_user_datadir()
        os.makedirs(spotilyzer_dir, exist_ok=True)
        if not os.path.exists(os.path.join(spotilyzer_dir, SEEDS_FILE)):
            seeds_file = os.path.join(get_package_datadir(), SEEDS_FILE)
            shutil.copy(seeds_file, spotilyzer_dir)



"""
spotilyzer seed JSON
"""

# system imports
import json

# 3rd party imports
import jsonschema

# schema
SEEDS_KEY = 'seeds'
GROUP_KEY = 'group'
CANDIDATES_KEY = 'candidates'
CANDIDATE_NAME_KEY = 'candidate-name'
INSTANCE_TYPES_KEY = 'instance-types'
_SCHEMA = {
    'type': 'object',
    'properties': {
        SEEDS_KEY: {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    GROUP_KEY: {
                        'type': 'string'
                    },
                    CANDIDATES_KEY: {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'properties': {
                                CANDIDATE_NAME_KEY: {
                                    'type': 'string'
                                },
                                INSTANCE_TYPES_KEY: {
                                    'type': 'array',
                                    'items': {
                                        'type': 'string',
                                        'pattern': '^[0-9a-z]+\\.[0-9a-z]+$'
                                    },
                                    'additionalItems': False,
                                    'uniqueItems': True,
                                    'minItems': 1
                                }
                            },
                            'required': [
                                CANDIDATE_NAME_KEY,
                                INSTANCE_TYPES_KEY
                            ],
                            'additionalProperties': False
                        },
                        'additionalItems': False,
                        'uniqueItems': True,
                        'minItems': 1
                    },
                },
                'required': [
                    GROUP_KEY,
                    CANDIDATES_KEY
                ],
                'additionalProperties': False
            },
            'additionalItems': False,
            'uniqueItems': True,
            'minItems': 1
        }
    },
    'required': [
        SEEDS_KEY
    ],
    'additionalProperties': False
}


def load_seeds(fseeds):
    """
    Load seeds JSON file. Perform schema and consistency checks.
    :param fseeds: path to JSON file
    :return: seeds object
    """
    with open(fseeds) as fobj:
        seeds = json.load(fobj)
    jsonschema.validate(seeds, _SCHEMA)
    _validate(seeds, fseeds)
    return seeds


def _validate(seeds, fseeds):
    _seeds = seeds[SEEDS_KEY]
    if len(_seeds) != len(set(s[GROUP_KEY] for s in _seeds)):
        raise SyntaxError(f"group names in {fseeds} are not unique")
    instance_types_set = set()
    num_instance_types = 0
    for seed in _seeds:
        candidates = seed[CANDIDATES_KEY]
        if len(candidates) != len(set(
            c[CANDIDATE_NAME_KEY] for c in candidates
        )):
            raise SyntaxError(f"candidate names for {seed[GROUP_KEY]} are not "
                              f"unique in {fseeds}")
        for candidate in candidates:
            instance_types = candidate[INSTANCE_TYPES_KEY]
            instance_types_set |= set(instance_types)
            num_instance_types += len(instance_types)
            if num_instance_types != len(instance_types_set):
                raise SyntaxError("duplicate instance type found in "
                                  f"{fseeds}: {seed[GROUP_KEY]}/"
                                  f"{candidate[CANDIDATE_NAME_KEY]}")



"""
spotilyzer group-requests subcommand
"""

# system imports
import math

# 3rd party imports
import tabulate

# project imports
from .base import SubCommand
from .utils.cli import options
from .json.candidates import load_candidates, GROUPS_KEY, GROUP_NAME_KEY, \
    MEM_CORE_RATIO_KEY
from .json.requests import load_requests as json_requests, save_requests, \
    REQUESTS_KEY, POD_NAME_KEY, MEM_LIMIT_KEY, CORE_LIMIT_KEY
from .csv.requests import load_requests as csv_requests

# constants
_DESCRIPTION = "place requested pods into node groups"
_FORMAT_OPT = 'format'
_JSON_FMT = 'json'
_CSV_FMT = 'csv'
_FORMAT_DEFAULT = _CSV_FMT
_CANDIDATES_ARG = 'candidates'
_REQUESTS_ARG = 'requests'
_OUTPUT_ARG = 'output'
_TABLE_HEADER = ('Pod', 'Group')


class GroupRequests(SubCommand):
    """
    group-requests subcommand
    """

    name = 'group-requests'

    @classmethod
    def add_parser(cls, subparsers):
        """
        Add group-requests subcommand parser.
        :param subparsers: object to attach parser
        :return: None
        """
        parser = subparsers.add_parser(cls.name, description=_DESCRIPTION,
                                       help=_DESCRIPTION)
        parser.add_argument(*options(_FORMAT_OPT),
                            choices=(_JSON_FMT, _CSV_FMT),
                            default=_FORMAT_DEFAULT,
                            help="request file format (default is "
                                 f"{_FORMAT_DEFAULT})")
        parser.add_argument(_CANDIDATES_ARG, help="JSON file with candidates")
        parser.add_argument(_REQUESTS_ARG, help="file containing requests")
        parser.add_argument(_OUTPUT_ARG,
                            help="output JSON file with grouped requests")

    def run(self):
        """
        Add groups to pod requests.
        :return: None
        """
        groups = load_candidates(self.getarg(_CANDIDATES_ARG))[GROUPS_KEY]
        requests = self._get_requests()
        results = self._group_requests(groups, requests)
        self._save_output(requests)
        self._display_results(results)

    def _get_requests(self):
        request_format = self.getarg(_FORMAT_OPT)
        if request_format == _CSV_FMT:
            return csv_requests(self.getarg(_REQUESTS_ARG))
        if request_format == _JSON_FMT:
            return json_requests(self.getarg(_REQUESTS_ARG))
        raise NotImplementedError(f"format {request_format} not implemented")

    @staticmethod
    def _group_requests(groups, requests):
        results = {}
        group_map = {g[GROUP_NAME_KEY]: g[MEM_CORE_RATIO_KEY] for g in groups}
        group_names = sorted(group_map.keys(),
                             key=lambda k: group_map[k])
        boundaries = [
            math.sqrt(group_map[group_names[i]] * group_map[group_names[i - 1]])
            for i in range(1, len(group_names))
        ]
        num_bdy = len(boundaries)
        for request in requests[REQUESTS_KEY]:
            mem_core_ratio = request[MEM_LIMIT_KEY] / request[CORE_LIMIT_KEY]
            found = False
            for i in range(num_bdy):
                if mem_core_ratio < boundaries[i]:
                    request[GROUP_NAME_KEY] = results[request[POD_NAME_KEY]] \
                        = group_names[i]
                    found = True
                    break
            if not found:
                request[GROUP_NAME_KEY] = results[request[POD_NAME_KEY]] \
                    = group_names[-1]
        return results

    def _save_output(self, requests):
        save_requests(self.getarg(_OUTPUT_ARG), requests)

    @staticmethod
    def _display_results(results):
        print()
        table = [[p, g] for p, g in results.items()]
        print(tabulate.tabulate(table, headers=_TABLE_HEADER))



"""
spotilyzer path utilities
"""

# system imports
import os
import pathlib

# exported constants
SEEDS_FILE = 'seeds.json'

# constants
_DATA_DIR = 'data'
_USER_DIR = '.spotilyzer'


def get_package_datadir():
    """
    Get the package data directory.
    :return: path to package data directory
    """
    return os.path.normpath(os.path.join(os.path.dirname(__file__),
                                         os.path.pardir, os.path.pardir,
                                         _DATA_DIR))


def get_user_datadir():
    """
    Get the user's data directory.
    :return: path to user's data directory
    """
    return os.path.join(pathlib.Path.home(), _USER_DIR)


def find_seeds_file():
    """
    Determine location of seeds.json using spotilyzer precedence. Raise
    exception if not found.
    :return: path to seeds.json
    """
    seeds_file = os.path.join(get_user_datadir(), SEEDS_FILE)
    if os.path.exists(seeds_file):
        return seeds_file
    seeds_file = os.path.join(get_package_datadir(), SEEDS_FILE)
    if os.path.exists(seeds_file):
        return seeds_file
    raise FileNotFoundError("seeds file not found")



"""
spotilyzer candidates JSON
"""

# system imports
import json

# 3rd party imports
import jsonschema

# project imports
from .seeds import CANDIDATES_KEY, CANDIDATE_NAME_KEY

# schema
REGION_KEY = 'region'
GROUPS_KEY = 'groups'
GROUP_NAME_KEY = 'group-name'
MEM_CORE_RATIO_KEY = 'mem-core-ratio'
PRICE_KEY = 'price'
RESOURCES_KEY = 'resources'
AVG_CORES_KEY = 'avg-cores'
AVG_MEMORY_KEY = 'avg-memory'
MIN_CORES_KEY = 'min-cores'
MIN_MEMORY_KEY = 'min-memory'
CANDIDATE_POOLS_KEY = 'pools'
INSTANCE_TYPE_KEY = 'instance-type'
AVAILABILITY_ZONES_KEY = 'availability-zones'
META_DATA_KEY = 'metadata'
TIMESTAMP_KEY = 'timestamp'
DURATION_KEY = 'duration'
MINPOOL_KEY = 'minpool'

_SCHEMA = {
    'type': 'object',
    'properties': {
        REGION_KEY: {
            'type': 'string'
        },
        GROUPS_KEY: {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    GROUP_NAME_KEY: {
                        'type': 'string'
                    },
                    MEM_CORE_RATIO_KEY: {
                        'type': 'number',
                        'exclusiveMinimum': 0.0
                    },
                    CANDIDATES_KEY: {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'properties': {
                                CANDIDATE_NAME_KEY: {
                                    'type': 'string'
                                },
                                PRICE_KEY: {
                                    'type': 'number',
                                    'minimum': 0.0
                                },
                                RESOURCES_KEY: {
                                    'type': 'object',
                                    'properties': {
                                        AVG_CORES_KEY: {
                                            'type': 'number',
                                            'exclusiveMinimum': 0.0
                                        },
                                        AVG_MEMORY_KEY: {
                                            'type': 'number',
                                            'exclusiveMinimum': 0.0
                                        },
                                        MIN_CORES_KEY: {
                                            'type': 'number',
                                            'minimum': 0.0
                                        },
                                        MIN_MEMORY_KEY: {
                                            'type': 'number',
                                            'minimum': 0.0
                                        }
                                    },
                                    'required': [
                                        AVG_CORES_KEY,
                                        AVG_MEMORY_KEY,
                                        MIN_CORES_KEY,
                                        MIN_MEMORY_KEY
                                    ],
                                    'additionalProperties': False
                                },
                                CANDIDATE_POOLS_KEY: {
                                    'type': 'array',
                                    'items': {
                                        'type': 'object',
                                        'properties': {
                                            INSTANCE_TYPE_KEY: {
                                                'type': 'string'
                                            },
                                            AVAILABILITY_ZONES_KEY: {
                                                'type': 'array',
                                                'items': {
                                                    'type': 'string'
                                                },
                                                'additionalItems': False,
                                                'uniqueItems': True,
                                                'minItems': 1
                                            }
                                        },
                                        'required': [
                                            INSTANCE_TYPE_KEY,
                                            AVAILABILITY_ZONES_KEY
                                        ],
                                        'additionalProperties': False
                                    },
                                    'additionalItems': False,
                                    'uniqueItems': True,
                                    'minItems': 1
                                }
                            },
                            'required': [
                                CANDIDATE_NAME_KEY,
                                PRICE_KEY,
                                RESOURCES_KEY,
                                CANDIDATE_POOLS_KEY
                            ],
                            'additionalProperties': False
                        },
                        'additionalItems': False,
                        'uniqueItems': True,
                        'minItems': 1
                    }
                },
                'required': [
                    GROUP_NAME_KEY,
                    MEM_CORE_RATIO_KEY,
                    CANDIDATES_KEY
                ],
                'additionalProperties': False
            },
            'additionalItems': False,
            'uniqueItems': True,
            'minItems': 1
        },
        META_DATA_KEY: {
            'type': 'object',
            'properties': {
                TIMESTAMP_KEY: {
                    'type': 'string',
                    'format': 'date-time'
                },
                DURATION_KEY: {
                    'type': 'integer',
                    'minimum': 1
                },
                MINPOOL_KEY: {
                    'type': 'integer',
                    'minimum': 1
                }
            },
            'required': [
                TIMESTAMP_KEY,
                DURATION_KEY,
                MINPOOL_KEY
            ],
            'additionalProperties': False
        }
    },
    'required': [
        REGION_KEY,
        GROUPS_KEY,
        META_DATA_KEY
    ],
    'additionalProperties': False
}

# constants
_JSON_INDENT = 2


def save_candidates(fcandidates, candidates):
    """
    Save candidates object to a JSON file. Schema and consistency checks are
    performed before saving.
    :param fcandidates: path to output file
    :param candidates: candidates object
    :return: None
    """
    jsonschema.validate(candidates, _SCHEMA)
    _validate(candidates, fcandidates)
    with open(fcandidates, 'w') as fobj:
        json.dump(candidates, fobj, indent=_JSON_INDENT)


def load_candidates(fcandidates):
    """
    Load candidates JSON file. Perform schema and consistency checks.
    :param fcandidates: path to JSON file
    :return: candidates object
    """
    with open(fcandidates) as fobj:
        candidates = json.load(fobj)
    jsonschema.validate(candidates, _SCHEMA)
    _validate(candidates, fcandidates)
    return candidates


def _validate(candidates, fname):
    groups = candidates[GROUPS_KEY]
    if len(groups) != len(set(g[GROUP_NAME_KEY] for g in groups)):
        raise SyntaxError(f"duplicate group name in {fname}")
    instance_types = set()
    num_instance_types = 0
    for group in groups:
        _candidates = group[CANDIDATES_KEY]
        if len(_candidates) != len(set(c[CANDIDATE_NAME_KEY]
                                       for c in _candidates)):
            raise SyntaxError("duplicate candidate name for group "
                              f"{group[GROUP_NAME_KEY]} in {fname}")
        for candidate in _candidates:
            add_set = set(p[INSTANCE_TYPE_KEY]
                          for p in candidate[CANDIDATE_POOLS_KEY])
            instance_types |= add_set
            num_instance_types += len(add_set)
            if num_instance_types != len(instance_types):
                raise SyntaxError("duplicate instance types for "
                                  f"{group[GROUP_NAME_KEY]}/"
                                  f"{candidate[CANDIDATE_NAME_KEY]} in "
                                  f"{fname}")



"""
spotilyzer subcommand base class
"""

# 3rd party imports
import boto3


class SubCommand:
    """
    Subcommand base class.
    """

    def __init__(self, args):
        self.client = boto3.client('ec2')
        self.args = args

    def reconnect(self, region):
        """
        Change region in client.
        :param region: AWS region name
        :return: None
        """
        if self.client.meta.region_name == region:
            return
        if region not in [r['RegionName']
                          for r in self.client.describe_regions()['Regions']]:
            raise ValueError(f"invalid region {region}")
        self.client = boto3.client('ec2', region_name=region)

    def getarg(self, name):
        """
        Return subcommand argument.
        :param name: argument name
        :return: argument value
        """
        return getattr(self.args, name.replace('-', '_'))



"""
spotilyzer driver
"""

# project imports
from .subcommands import subcommands
from .cli import SUBCOMMAND_ARG


def run(args):
    """
    Execute subcommand.
    :param args: argparse Namespace
    :return: None
    """
    name = getattr(args, SUBCOMMAND_ARG)
    for subcommand in subcommands:
        if name == subcommand.name:
            subcommand(args).run()
            return
    raise NotImplementedError(f"subcommand {name} not implemented")



"""
spotilyzer subcommands entry point
"""

# project imports
from .init import Init
from .create_candidates import CreateCandidates
from .group_requests import GroupRequests
from .size_fleets import SizeFleets

# exports
subcommands = [Init, CreateCandidates, GroupRequests, SizeFleets]
